import numpy as np
import pandas as pd
import polars as pl
import pytest
from statsforecast.utils import generate_series
from utilsforecast.data import generate_series as util_generate_series

from hierarchicalforecast.utils import (
    CodeTimer,
    _ma_cov,
    _shrunk_covariance_schaferstrimmer_no_nans,
    _shrunk_covariance_schaferstrimmer_with_nans,
    aggregate,
    aggregate_temporal,
    get_cross_temporal_tags,
    level_to_outputs,
    make_future_dataframe,
    quantiles_to_outputs,
    samples_to_quantiles_df,
)

from .conftest import assert_raises_with_message


@pytest.fixture
def df():
    df = pd.DataFrame(
        {
            'cat1': ['a', 'a', 'a', 'b'],
            'cat2': ['1', '2', '3', '2'],
            'y': [10, 20, 30, 40],
            'ds': ['2020-01-01', '2020-02-01', '2020-03-01', '2020-02-01']
        }
    )
    df['country'] = 'COUNTRY'
    return df

@pytest.fixture
def df_pl(df):
    return pl.from_pandas(df)

@pytest.fixture
def spec():
    spec = [['country'], ['country', 'cat1'], ['country', 'cat1', 'cat2']]
    return spec

# simple case
def test_simple_case(df, spec):
    Y_df, S_df, tags = aggregate(df, spec)
    assert list(Y_df["unique_id"]) == (3 * ['COUNTRY'] + 3 * ['COUNTRY/a'] + ['COUNTRY/b'] + ['COUNTRY/a/1', 'COUNTRY/a/2', 'COUNTRY/a/3'] + ['COUNTRY/b/2'])

    assert Y_df.query("unique_id == 'COUNTRY'")['y'].tolist() == [10, 60, 30]
    assert list(S_df["unique_id"]) == ['COUNTRY', 'COUNTRY/a', 'COUNTRY/b', 'COUNTRY/a/1', 'COUNTRY/a/2', 'COUNTRY/a/3', 'COUNTRY/b/2']

    assert S_df.columns.tolist() == ['unique_id', 'COUNTRY/a/1', 'COUNTRY/a/2', 'COUNTRY/a/3', 'COUNTRY/b/2']

    expected_tags = {
        'country': ['COUNTRY'],
        'country/cat1': ['COUNTRY/a', 'COUNTRY/b'],
        'country/cat1/cat2': ['COUNTRY/a/1', 'COUNTRY/a/2', 'COUNTRY/a/3','COUNTRY/b/2'],
    }
    for k, actual in tags.items():
        assert actual.tolist() == expected_tags[k]

    # test categoricals don't produce all combinations
    df2 = df.copy()
    for col in ('country', 'cat1', 'cat2'):
        df2[col] = df2[col].astype('category')

    Y_df2, *_ = aggregate(df2, spec)
    assert Y_df.shape[0] == Y_df2.shape[0]


# Testing equivalence of other dataframe libs to pandas results.
# TODO: extend for other frameworks
def assert_eq_agg_dataframe(df, spec, frameworks=None):
    if frameworks is None:
        frameworks = ["polars"]
    for framework in frameworks:
        if framework == "polars":
            df_f = pl.from_pandas(df)
        else:
            raise ValueError(f"Framework {framework} not recognized")

        Y_df, S_df, tags = aggregate(df=df, spec=spec)
        Y_df_f, S_df_f, tags_f = aggregate(df=df_f, spec=spec)

        pd.testing.assert_frame_equal(Y_df, Y_df_f.to_pandas())
        pd.testing.assert_frame_equal(S_df, S_df_f.to_pandas())
        for tag in tags:
            np.testing.assert_equal(tags[tag], tags_f[tag])


def test_unbalanced_dataset():
    # test unbalanced dataset
    max_tenure = 24
    dates = pd.date_range(start='2019-01-31', freq='ME', periods=max_tenure)
    cohort_tenure = [24, 23, 22, 21]

    ts_list = []

    # Create ts for each cohort
    for i in range(len(cohort_tenure)):
        ts_list.append(
            generate_series(n_series=1, freq='M', min_length=cohort_tenure[i], max_length=cohort_tenure[i]).reset_index() \
                .assign(ult=i) \
                .assign(ds=dates[-cohort_tenure[i]:]) \
                .drop(columns=['unique_id'])
        )
    df = pd.concat(ts_list, ignore_index=True)

    # Create categories
    df['pen'] = np.where(df['ult'] < 2, 'a', 'b')
    # Note that unique id requires strings
    df['ult'] = df['ult'].astype(str)

    hier_levels = [
        ['pen'],
        ['pen', 'ult'],
    ]
    assert_eq_agg_dataframe(df, hier_levels)


def test_tourism_df_non_null_grouped(tourism_df, hiers_grouped, hiers_strictly):
    df = tourism_df
    # test strict
    hier_df, S_df, tags = aggregate(df=df, spec=hiers_strictly)
    assert len(hier_df) == 6800
    assert hier_df["unique_id"].nunique(), 85
    assert S_df.shape, (85, 77)
    np.testing.assert_array_equal(hier_df["unique_id"].unique(), S_df["unique_id"])
    assert len(tags), len(hiers_strictly)


    # test grouped
    hier_df, S_df, tags = aggregate(df=df, spec=hiers_grouped)
    assert len(hier_df) == 34_000
    assert hier_df["unique_id"].nunique(), 425
    assert S_df.shape == (425, 305)
    np.testing.assert_array_equal(hier_df["unique_id"].unique(), S_df["unique_id"])
    assert len(tags) == len(hiers_grouped)
    # polars
    assert_eq_agg_dataframe(df, hiers_strictly)
    assert_eq_agg_dataframe(df, hiers_grouped)

def test_tourism_df_null_grouped(tourism_df, hiers_strictly):
    df = tourism_df
    #Unit Test NaN Values
    df_nan = df.copy()
    df_nan.loc[0, 'Region'] = float('nan')
    assert_raises_with_message(
        aggregate,
        'null values',
        df_nan, hiers_strictly,
    )

    #Unit Test None Values
    df_none = df.copy()
    df_none.loc[0, 'Region'] = None
    assert_raises_with_message(
        aggregate,
        'null values',
        df_none, hiers_strictly,
    )
    # polars

    #Unit Test NaN Values
    df_nan_pl = pl.DataFrame(df_nan)
    assert_raises_with_message(
        aggregate,
        'null values',
        df_nan_pl, hiers_strictly,
    )

    #Unit Test None Values
    df_none_pl = pl.DataFrame(df_none)
    assert_raises_with_message(
        aggregate,
        'null values',
        df_none_pl, hiers_strictly,
    )

def test_equality_sparse_non_sparse(tourism_df, hiers_strictly, hiers_grouped):
    df = tourism_df

    # Test equality of sparse and non-sparse aggregation
    with CodeTimer('strict non-sparse aggregate'):
        Y_df, S_df, tags = aggregate(df=df, sparse_s=False, spec=hiers_strictly)

    with CodeTimer('strict sparse aggregate'):
        Y_df_sparse, S_df_sparse, tags_sparse = aggregate(df=df, sparse_s=True, spec=hiers_strictly)

    np.testing.assert_almost_equal(Y_df.y.values, Y_df_sparse.y.values)
    np.testing.assert_array_equal(S_df.values, S_df_sparse.values)

    np.testing.assert_array_equal(S_df.columns, S_df_sparse.columns)
    np.testing.assert_array_equal(S_df.index, S_df_sparse.index)

    np.testing.assert_array_equal(Y_df.columns, Y_df_sparse.columns)
    np.testing.assert_array_equal(Y_df.index, Y_df_sparse.index)

    with CodeTimer('grouped non-sparse aggregate'):
        Y_df, S_df, tags = aggregate(df=df, sparse_s=False, spec=hiers_grouped)

    with CodeTimer('grouped sparse aggregate'):
        Y_df_sparse, S_df_sparse, tags_sparse = aggregate(df=df, sparse_s=True, spec=hiers_grouped)

    np.testing.assert_almost_equal(Y_df.y.values, Y_df_sparse.y.values)
    np.testing.assert_array_equal(S_df.values, S_df_sparse.values)

    np.testing.assert_array_equal(S_df.columns, S_df_sparse.columns)
    np.testing.assert_array_equal(S_df.index, S_df_sparse.index)

    np.testing.assert_array_equal(Y_df.columns, Y_df_sparse.columns)
    np.testing.assert_array_equal(Y_df.index, Y_df_sparse.index)


@pytest.fixture
def df_for_other_cases():
    df = pd.DataFrame(
        {
            'cat1': 8 * ['a'] + 8 * ['b'],
            'y': np.linspace(10, 160, 16),
            'ds': 2 * ['2020-01-01', '2020-04-01', '2020-07-01', '2020-10-01', '2021-01-01', '2021-04-01', '2021-07-01', '2021-10-01']
        }
    )
    df['country'] = 'COUNTRY'
    df["ds"] = pd.to_datetime(df["ds"])
    df["unique_id"] = df["country"] + "/" + df["cat1"]
    return df

def test_other_simple_case(df_for_other_cases):
    df = df_for_other_cases
    # simple case
    spec = {"year": 4,
            "semiannual": 2,
            "quarter": 1}

    Y_df_te, S_df_te, tags_te = aggregate_temporal(df, spec)

    np.testing.assert_array_equal(
        list(Y_df_te["unique_id"]),
            ['COUNTRY/a', 'COUNTRY/a', 'COUNTRY/b', 'COUNTRY/b', 'COUNTRY/a',
        'COUNTRY/a', 'COUNTRY/a', 'COUNTRY/a', 'COUNTRY/b', 'COUNTRY/b',
        'COUNTRY/b', 'COUNTRY/b', 'COUNTRY/a', 'COUNTRY/a', 'COUNTRY/a',
        'COUNTRY/a', 'COUNTRY/a', 'COUNTRY/a', 'COUNTRY/a', 'COUNTRY/a',
        'COUNTRY/b', 'COUNTRY/b', 'COUNTRY/b', 'COUNTRY/b', 'COUNTRY/b',
        'COUNTRY/b', 'COUNTRY/b', 'COUNTRY/b'])
    np.testing.assert_array_equal(
        list(Y_df_te["temporal_id"]),
            ['year-1', 'year-2', 'year-1', 'year-2', 'semiannual-1', 'semiannual-2', 'semiannual-3', 'semiannual-4', 'semiannual-1', 'semiannual-2', 'semiannual-3', 'semiannual-4', 'quarter-1', 'quarter-2', 'quarter-3', 'quarter-4', 'quarter-5', 'quarter-6', 'quarter-7', 'quarter-8', 'quarter-1', 'quarter-2', 'quarter-3', 'quarter-4', 'quarter-5', 'quarter-6', 'quarter-7', 'quarter-8']
    )
    np.testing.assert_array_equal(Y_df_te.query("unique_id == 'COUNTRY/a' & temporal_id in ['year-1', 'year-2']")['y'].values, [100, 260])
    np.testing.assert_array_equal(
        list(S_df_te["temporal_id"]),
        ['year-1', 'year-2', 'semiannual-1', 'semiannual-2', 'semiannual-3', 'semiannual-4', 'quarter-1', 'quarter-2', 'quarter-3', 'quarter-4',
    'quarter-5', 'quarter-6', 'quarter-7', 'quarter-8'],
    )
    np.testing.assert_array_equal( S_df_te.columns,
        ['temporal_id', 'quarter-1',
        'quarter-2', 'quarter-3',
        'quarter-4', 'quarter-5',
        'quarter-6', 'quarter-7',
        'quarter-8'],
    )
    expected_tags = {
        'year': ['year-1', 'year-2'],
        'semiannual': ['semiannual-1', 'semiannual-2', 'semiannual-3', 'semiannual-4'],
        'quarter': ['quarter-1', 'quarter-2', 'quarter-3', 'quarter-4', 'quarter-5', 'quarter-6', 'quarter-7', 'quarter-8'],
    }
    for k, actual in tags_te.items():
        np.testing.assert_array_equal(actual, expected_tags[k])


def test_future_make_df(df_for_other_cases):
    df = df_for_other_cases
    # Test make_future_dataframe
    futr_df = make_future_dataframe(df, freq='QS', h=4)

    futr_df_check = pd.DataFrame(
        {
            'unique_id': 4 * ['COUNTRY/a'] + 4 *['COUNTRY/b'],
            'ds': pd.concat(2 * [pd.date_range(start='2022-01-01', freq='QS', periods=4).to_series()], ignore_index=True),
        }
    )

    pd.testing.assert_frame_equal(futr_df, futr_df_check)
    # polars
    df_pl = pl.DataFrame(df)
    futr_df_pl = make_future_dataframe(df_pl, freq='1q', h=4)
    pd.testing.assert_frame_equal(futr_df, futr_df_pl.to_pandas())

def test_cross_temporal_tags():
    # Test cross-temporal tags
    df = pd.DataFrame(
        {
            'cat1': 8 * ['a'] + 8 * ['b'],
            'y': np.linspace(10, 160, 16),
            'ds': 2 * ['2020-01-01', '2020-04-01', '2020-07-01', '2020-10-01', '2021-01-01', '2021-04-01', '2021-07-01', '2021-10-01']
        }
    )
    df['country'] = 'COUNTRY'
    df["ds"] = pd.to_datetime(df["ds"])

    # Aggregate cross-sectionally
    spec_cs =  [['country'], ['country', 'cat1']]
    Y_df_cs, S_df_cs, tags_cs = aggregate(df, spec_cs)

    # Aggregate temporally
    spec_te = {"year": 4, "quarter": 1}
    Y_df_te, S_df_te, tags_te = aggregate_temporal(Y_df_cs, spec_te)

    # Get cross-temporal tags
    cross_df, tags_ct = get_cross_temporal_tags(Y_df_te, tags_cs, tags_te)

    tags_ct_check = {'country//year': ['COUNTRY//year-1', 'COUNTRY//year-2'],
    'country//quarter': ['COUNTRY//quarter-1',
    'COUNTRY//quarter-2',
    'COUNTRY//quarter-3',
    'COUNTRY//quarter-4',
    'COUNTRY//quarter-5',
    'COUNTRY//quarter-6',
    'COUNTRY//quarter-7',
    'COUNTRY//quarter-8'],
    'country/cat1//year': ['COUNTRY/a//year-1',
    'COUNTRY/a//year-2',
    'COUNTRY/b//year-1',
    'COUNTRY/b//year-2'],
    'country/cat1//quarter': ['COUNTRY/a//quarter-1',
    'COUNTRY/a//quarter-2',
    'COUNTRY/a//quarter-3',
    'COUNTRY/a//quarter-4',
    'COUNTRY/a//quarter-5',
    'COUNTRY/a//quarter-6',
    'COUNTRY/a//quarter-7',
    'COUNTRY/a//quarter-8',
    'COUNTRY/b//quarter-1',
    'COUNTRY/b//quarter-2',
    'COUNTRY/b//quarter-3',
    'COUNTRY/b//quarter-4',
    'COUNTRY/b//quarter-5',
    'COUNTRY/b//quarter-6',
    'COUNTRY/b//quarter-7',
    'COUNTRY/b//quarter-8']}

    for key, value in tags_ct.items():
        np.testing.assert_array_equal(value, tags_ct_check[key])
    # polars
    df_pl = pl.DataFrame(df)
    Y_df_cs_pl, S_df_cs_pl, tags_cs_pl = aggregate(df_pl, spec_cs)
    Y_df_te_pl, S_df_te_pl, tags_te_pl = aggregate_temporal(Y_df_cs_pl, spec_te)
    cross_df_pl, tags_ct_pl = get_cross_temporal_tags(Y_df_te_pl, tags_cs_pl, tags_te_pl)

    for key, value in tags_ct_pl.items():
        np.testing.assert_array_equal(value, tags_ct_check[key])


def test_temporal_agg():
    # Test temporal aggregation with synthetic series of varying length
    freq1 = "D"
    df1 = util_generate_series(n_series=2, freq=freq1, min_length=2 * 365, max_length=4 * 365, static_as_categorical=False, equal_ends=True)

    spec1  = {"year": 365, "quarter": 91, "month": 30, "week": 7, "day": 1}

    Y_df1_te, S_df1_te, tags1_te = aggregate_temporal(df1, spec1)

    assert Y_df1_te.isnull().sum().sum() == 0, "There are NaN values in the aggregated DataFrame"

    freq2 = "ME"
    df2 = util_generate_series(n_series=2, freq=freq2, min_length=24, max_length=48, static_as_categorical=False, equal_ends=True)

    spec2  = {"year": 12, "quarter": 4, "month": 1}

    Y_df2_te, S_df2_te, tags2_te = aggregate_temporal(df2, spec2)

    assert Y_df2_te.isnull().sum().sum() == 0, "There are NaN values in the aggregated DataFrame"
    # polars
    df1_pl = pl.from_pandas(df1)
    df2_pl = pl.from_pandas(df2)

    Y_df1_te_pl, S_df1_te_pl, tags1_te_pl = aggregate_temporal(df1_pl, spec1)
    Y_df2_te_pl, S_df2_te_pl, tags2_te_pl = aggregate_temporal(df2_pl, spec2)

    # Polars adds nanoseconds to the datetime columns, so we remove them for comparison
    Y_df1_te_pl_pd = Y_df1_te_pl.to_pandas()
    Y_df2_te_pl_pd = Y_df2_te_pl.to_pandas()

    pd.testing.assert_frame_equal(Y_df1_te, Y_df1_te_pl_pd)
    pd.testing.assert_frame_equal(Y_df2_te, Y_df2_te_pl_pd)


@pytest.mark.parametrize("levels, expected_quantiles, expected_names", [
    ([80, 90], [0.5, 0.05, 0.1, 0.9, 0.95], ['-median', '-lo-90', '-lo-80', '-hi-80', '-hi-90']),
    ([30], [0.5, 0.35, 0.65], ['-median', '-lo-30', '-hi-30'])
])
def test_level_to_outputs(levels, expected_quantiles, expected_names):
    result_quantiles, result_names = level_to_outputs(levels)
    np.testing.assert_array_equal(result_quantiles, expected_quantiles)
    np.testing.assert_array_equal(result_names, expected_names)


@pytest.mark.parametrize("quantiles, expected_quantiles, expected_names", [
    ([0.2, 0.4, 0.6, 0.8], [0.2, 0.4, 0.6, 0.8], ['-lo-60.0', '-lo-20.0', '-hi-20.0', '-hi-60.0']),
    ([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9], [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9], ['-lo-80.0', '-lo-60.0', '-lo-40.0', '-lo-20.0', '-median', '-hi-20.0', '-hi-40.0', '-hi-60.0', '-hi-80.0'])
])
def test_quantiles_to_outputs(quantiles, expected_quantiles, expected_names):
    result_quantiles, result_names = quantiles_to_outputs(quantiles)
    np.testing.assert_array_equal(result_quantiles, expected_quantiles)
    np.testing.assert_array_equal(result_names, expected_names)


def test_samples_to_quantiles_df():
    #samples_to_quantiles_df unit tests
    start_date = pd.Timestamp("2023-06-01")
    end_date = pd.Timestamp("2023-06-10")
    frequency = "D"  # Daily frequency
    dates = pd.date_range(start=start_date, end=end_date, freq=frequency).tolist()
    samples = np.random.rand(3, 200, 10)
    unique_ids = ['id1', 'id2', 'id3']
    level = np.array([10, 50, 90])
    quantiles=np.array([0.5, 0.05, 0.25, 0.45, 0.55, 0.75, 0.95])

    ret_quantiles_1, ret_df_1 = samples_to_quantiles_df(samples, unique_ids, dates, level=level)
    ret_quantiles_2, ret_df_2 = samples_to_quantiles_df(samples, unique_ids, dates, quantiles=quantiles)

    np.testing.assert_array_equal(ret_quantiles_1, quantiles)
    np.testing.assert_array_equal(ret_df_1.columns,
        ['unique_id', 'ds', 'model', 'model-median', 'model-lo-90', 'model-lo-50', 'model-lo-10', 'model-hi-10', 'model-hi-50', 'model-hi-90']
    )
    np.testing.assert_array_equal(ret_df_1["unique_id"].values,
        ['id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1',
        'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2',
        'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3']
    )
    np.testing.assert_array_equal(ret_quantiles_1, ret_quantiles_2)
    np.testing.assert_array_equal(ret_df_1["unique_id"], ret_df_2["unique_id"])
    # polars

    ret_quantiles_1, ret_df_1 = samples_to_quantiles_df(samples, unique_ids, dates, level=level, backend='polars')
    ret_quantiles_2, ret_df_2 = samples_to_quantiles_df(samples, unique_ids, dates, quantiles=quantiles, backend='polars')

    np.testing.assert_array_equal(ret_quantiles_1, quantiles)
    np.testing.assert_array_equal(ret_df_1.columns,
        ['unique_id', 'ds', 'model', 'model-median', 'model-lo-90', 'model-lo-50', 'model-lo-10', 'model-hi-10', 'model-hi-50', 'model-hi-90']
    )
    np.testing.assert_array_equal(
        list(ret_df_1["unique_id"]),
        ['id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1', 'id1',
        'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2', 'id2',
        'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3', 'id3']
    )
    np.testing.assert_array_equal(ret_quantiles_1, ret_quantiles_2)
    np.testing.assert_array_equal(ret_df_1["unique_id"], ret_df_2["unique_id"])

def test_agg_func_for_exog_vars():
    #Unit test for the aggregate function accounting for exog_vars
    df = pd.DataFrame(
        {
            'cat1': ['a', 'a', 'c'],
            'cat2': ['1', '2', '3'],
            'exog1': [4, 5, 6],
            'exog2': [7, 6, 5],
            'y': [10, 20, 30],
            'ds': ['2020-01-01', '2020-02-01', '2020-03-01']
        }
    )
    spec = [["cat1"],["cat1","cat2"]]


    Y_df_check = pd.DataFrame(
        data={
            'unique_id': ['a', 'a', 'c', 'a/1', 'a/2', 'c/3'],
            'ds': ['2020-01-01','2020-02-01','2020-03-01','2020-01-01','2020-02-01','2020-03-01'],
            'y': [10, 20, 30, 10, 20, 30],
        },
    )

    S_df_check = pd.DataFrame(
        data={
            'unique_id': ['a','c','a/1','a/2','c/3'],
            'a/1': np.array([1.0, 0.0, 1.0, 0.0, 0.0], dtype=np.float64),
            'a/2': np.array([1.0, 0.0, 0.0, 1.0, 0.0], dtype=np.float64),
            'c/3': np.array([0.0, 1.0, 0.0, 0.0, 1.0], dtype=np.float64)
        },
    )


    Y_df_check_exog = pd.DataFrame(
        data = {
            'unique_id': ['a', 'a', 'c', 'a/1', 'a/2', 'c/3'],
            'ds': ['2020-01-01', '2020-02-01', '2020-03-01', '2020-01-01', '2020-02-01', '2020-03-01'],
            'y': [10, 20, 30, 10, 20, 30],
            'exog1_mean': [4.0, 5.0, 6.0, 4.0, 5.0, 6.0],
            'exog2_sum': [7, 6, 5, 7, 6, 5]
        },
    )

    Y_df, S_df, tags = aggregate(
        df = df,
        spec = spec,
        exog_vars = None,
    )

    Y_df_exog, S_df_exog, tags = aggregate(
        df = df,
        spec = spec,
        exog_vars = {'exog1':'mean','exog2':'sum'},
    )

    np.testing.assert_array_equal(Y_df, Y_df_check)

    np.testing.assert_array_equal(S_df, S_df_check)

    np.testing.assert_array_equal(Y_df_exog, Y_df_check_exog)
    # polars

    df_f = pl.from_pandas(df)
    Y_df_check_f = pl.from_pandas(Y_df_check)
    S_df_check_f = pl.from_pandas(S_df_check)
    Y_df_check_exog_f = pl.from_pandas(Y_df_check_exog)

    Y_df, S_df, tags = aggregate(
        df = df_f,
        spec = spec,
        exog_vars = None,
    )

    Y_df_exog, S_df_exog, tags = aggregate(
        df = df_f,
        spec = spec,
        exog_vars = {'exog1':'mean','exog2':'sum'},
    )

    np.testing.assert_array_equal(Y_df, Y_df_check_f)

    np.testing.assert_array_equal(S_df, S_df_check_f)

    np.testing.assert_array_equal(Y_df_exog, Y_df_check_exog_f)

def test_cov_equivalence():
    # test covariance equivalence
    n_samples = 100
    n_hiers = 10
    y_insample = np.random.rand(n_samples, n_hiers)
    y_hat_insample = np.random.rand(n_samples, n_hiers)
    residuals = (y_insample - y_hat_insample)
    nan_mask = np.isnan(residuals)

    # Check equivalence of covariance functions in case of no nans
    W_nb = _ma_cov(residuals, ~nan_mask)
    W_np = np.cov(residuals)
    np.testing.assert_allclose(W_nb, W_np, atol=1e-6)

    # Check equivalence of shrunk covariance functions in case of no nans
    W_ss_nonan = _shrunk_covariance_schaferstrimmer_no_nans(residuals, 2e-8)
    W_ss_nan = _shrunk_covariance_schaferstrimmer_with_nans(residuals, ~nan_mask, 2e-8)
    np.testing.assert_allclose(W_ss_nan, W_ss_nonan, atol=1e-6)

    # Check equivalence of diagonal elements of shrunk W to non-shrunk W in case of no nans
    np.testing.assert_allclose(np.diag(W_np), np.diag(W_ss_nan), atol=1e-6)



from dataclasses import dataclass

import numpy as np
import pytest
from scipy import sparse

from hierarchicalforecast.methods import (
    ERM,
    BottomUp,
    MiddleOut,
    MiddleOutSparse,
    MinTrace,
    MinTraceSparse,
    OptimalCombination,
    TopDown,
    TopDownSparse,
    _is_strictly_hierarchical,
    is_strictly_hierarchical,
)
from hierarchicalforecast.utils import _construct_adjacency_matrix


@dataclass
class HierarchicalTestData:
    """Test data for hierarchical forecasting methods."""
    S: np.ndarray  # Summing matrix
    h: int  # Forecast horizon
    y_bottom: np.ndarray  # Bottom level historical data
    y_hat_bottom: np.ndarray  # Bottom level forecasts
    y_hat_bottom_insample: np.ndarray  # Bottom level insample forecasts
    tags: dict[str, np.ndarray]  # Hierarchy level tags

@pytest.fixture
def hierarchical_data():
    """Fixture providing test data for hierarchical forecasting methods."""
    # Summing matrix for 3-level hierarchy (7x4)
    # Level 1: Total (1 series)
    # Level 2: Two groups (2 series)
    # Level 3: Four bottom series (4 series)
    S = np.array([
        [1.0, 1.0, 1.0, 1.0],  # Total
        [1.0, 1.0, 0.0, 0.0],  # Group 1
        [0.0, 0.0, 1.0, 1.0],  # Group 2
        [1.0, 0.0, 0.0, 0.0],  # Bottom 1
        [0.0, 1.0, 0.0, 0.0],  # Bottom 2
        [0.0, 0.0, 1.0, 0.0],  # Bottom 3
        [0.0, 0.0, 0.0, 1.0],  # Bottom 4
    ])

    h = 2  # Forecast horizon
    base_series = np.array([10.0, 5.0, 4.0, 2.0, 1.0])

    # Create bottom level historical data (4 series x 5 time periods)
    y_bottom = np.vstack([i * base_series for i in range(1, 5)])

    # Create insample forecasts with NaN in first period
    y_hat_bottom_insample = np.roll(y_bottom, 1, axis=1)
    y_hat_bottom_insample[:, 0] = np.nan

    # Create simple forecasts for bottom series
    y_hat_bottom = np.vstack([i * np.ones(h) for i in range(1, 5)])

    tags = {
        "level1": np.array([0]),
        "level2": np.array([1, 2]),
        "level3": np.array([3, 4, 5, 6])
    }

    return HierarchicalTestData(
        S=S,
        h=h,
        y_bottom=y_bottom,
        y_hat_bottom=y_hat_bottom,
        y_hat_bottom_insample=y_hat_bottom_insample,
        tags=tags
    )

def test_sigmah_hierarchy(hierarchical_data):
    """Test bottom up forecast recovery matches expected output."""
    data = hierarchical_data
    cls_bottom_up = BottomUp()
    result = cls_bottom_up(S=data.S, y_hat=data.S @ data.y_hat_bottom)["mean"]
    expected = data.S @ data.y_hat_bottom
    assert result.all() == expected.all()


def test_btm_up_forecast_recovery(hierarchical_data):
    """Test bottom up forecast recovery."""
    data = hierarchical_data
    cls_bottom_up = BottomUp()
    bu_bootstrap_intervals = cls_bottom_up(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
    )
    expected = data.S @ data.y_hat_bottom
    assert bu_bootstrap_intervals["mean"].all() == expected.all()


def test_forecast_recovery_fit_predict(hierarchical_data):
    """Test forecast recovery with fit -> predict pattern."""
    data = hierarchical_data
    cls_bottom_up = BottomUp()
    cls_bottom_up.fit(S=data.S, y_hat=data.S @ data.y_hat_bottom)
    y_tilde = cls_bottom_up.predict(S=data.S, y_hat=data.S @ data.y_hat_bottom)["mean"]
    expected = data.S @ data.y_hat_bottom
    assert y_tilde.all() == expected.all()


def test_not_fitted(hierarchical_data):
    """Test not fitted message for unfitted predict."""
    data = hierarchical_data
    cls_bottom_up = BottomUp()
    with pytest.raises(Exception) as exec_info:
        cls_bottom_up.predict(data.S, data.S @ data.y_hat_bottom)
    assert "not fitted yet" in str(exec_info.value)


@pytest.fixture
def tags_non_hier():
    return {
        "Country": np.array([0]),
        "Country/City": np.array([2, 1]),
        "Country/Transgender": np.array([3, 4]),
        "Country-City-Transgender": np.array([5, 6, 7, 8]),
    }

@pytest.fixture
def S_non_hier():
    return np.array(
            [
                [1.0, 1.0, 1.0, 1.0],  # total
                [1.0, 1.0, 0.0, 0.0],  # city 1
                [0.0, 0.0, 1.0, 1.0],  # city 2
                [1.0, 0.0, 1.0, 0.0],  # transgender
                [0.0, 1.0, 0.0, 1.0],  # no transgender
                [1.0, 0.0, 0.0, 0.0],  # city 1 - transgender
                [0.0, 1.0, 0.0, 0.0],  # city 1 - no transgender
                [0.0, 0.0, 1.0, 0.0],  # city 2 - transgender
                [0.0, 0.0, 0.0, 1.0],  # city 2 - no transgender
            ]
        )


def test_is_strictly_hierarchical(hierarchical_data, S_non_hier, tags_non_hier):
    data = hierarchical_data
    assert is_strictly_hierarchical(data.S, data.tags)
    assert not is_strictly_hierarchical(S_non_hier, tags_non_hier)


def test_hidden_is_strictly_hierarchical(hierarchical_data, S_non_hier, tags_non_hier):
    data = hierarchical_data
    A = np.array(
        [
            [0, 1, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 1],
        ]
    )
    assert _construct_adjacency_matrix(sparse.csr_matrix(data.S), data.tags).toarray().all() == A.all()
    assert _is_strictly_hierarchical(sparse.csr_matrix(A, dtype=bool), data.tags)
    A_non_hier = np.array(
        [
            [0, 1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 1],
        ]
    )
    assert _construct_adjacency_matrix(sparse.csr_matrix(S_non_hier), tags_non_hier).toarray().all() == A_non_hier.all()
    assert not _is_strictly_hierarchical(sparse.csr_matrix(A_non_hier, dtype=bool), tags_non_hier)


@pytest.mark.parametrize("method", ["forecast_proportions", "average_proportions", "proportion_averages"])
def test_top_down_forecast_recovery(hierarchical_data, method):
    """Test TopDown methods can recover forecasts when time series share same proportions."""
    data = hierarchical_data

    cls_top_down = TopDown(method=method)
    if cls_top_down.insample:
        assert method in ["average_proportions", "proportion_averages"]
        result = cls_top_down(
            S=data.S, y_hat=data.S @ data.y_hat_bottom, y_insample=data.S @ data.y_bottom, tags=data.tags
        )["mean"]
    else:
        result = cls_top_down(S=data.S, y_hat=data.S @ data.y_hat_bottom, tags=data.tags)["mean"]

    np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


def test_top_down_sparse_hierarchical_error(S_non_hier, tags_non_hier):
    """Test TopDownSparse raises error for non-hierarchical structures."""
    cls_top_down = TopDownSparse(method="average_proportions")
    # Create dummy data
    y_hat = np.ones((S_non_hier.shape[0], 2))
    y_insample = np.ones((S_non_hier.shape[0], 5))
    with pytest.raises(Exception, match="Top-down reconciliation requires strictly hierarchical structures"):
        cls_top_down(S=sparse.csr_matrix(S_non_hier), y_hat=y_hat, y_insample=y_insample, tags=tags_non_hier)


@pytest.mark.parametrize("method", ["forecast_proportions", "average_proportions", "proportion_averages"])
def test_top_down_sparse_forecast_recovery(hierarchical_data, method):
    """Test TopDownSparse methods can recover forecasts."""
    data = hierarchical_data

    cls_top_down = TopDownSparse(method=method)
    if cls_top_down.insample:
        assert method in ["average_proportions", "proportion_averages"]
        result = cls_top_down(
            S=sparse.csr_matrix(data.S),
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
    else:
        result = cls_top_down(S=sparse.csr_matrix(data.S), y_hat=data.S @ data.y_hat_bottom, tags=data.tags)["mean"]

    np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


@pytest.mark.parametrize("method", ["forecast_proportions", "average_proportions", "proportion_averages"])
def test_top_down_vs_sparse_equivalence(hierarchical_data, method):
    """Test TopDown and TopDownSparse produce equivalent results."""
    data = hierarchical_data

    cls_top_down = TopDown(method=method)
    cls_top_down_sparse = TopDownSparse(method=method)

    if cls_top_down.insample:
        assert method in ["average_proportions", "proportion_averages"]
        result_dense = cls_top_down(
            S=data.S, y_hat=data.S @ data.y_hat_bottom, y_insample=data.S @ data.y_bottom, tags=data.tags
        )["mean"]
        result_sparse = cls_top_down_sparse(
            S=sparse.csr_matrix(data.S),
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
        tolerance = np.finfo(np.float64).eps
    else:
        result_dense = cls_top_down(S=data.S, y_hat=data.S @ data.y_hat_bottom, tags=data.tags)["mean"]
        result_sparse = cls_top_down_sparse(
            S=sparse.csr_matrix(data.S),
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
        tolerance = 1e-9

    np.testing.assert_allclose(result_dense, result_sparse, tolerance)


@pytest.mark.parametrize("method", ["average_proportions", "proportion_averages"])
def test_top_down_nan_handling(hierarchical_data, method):
    """Test TopDown handles NaN values correctly in insample data."""
    data = hierarchical_data

    cls_top_down = TopDown(method=method)

    # Original result with zeros
    y_insample_orig = data.S @ data.y_bottom
    y_insample_orig[-1, :] = 0
    result_orig = cls_top_down(
        S=data.S, y_hat=data.S @ data.y_hat_bottom, y_insample=y_insample_orig, tags=data.tags
    )["mean"]

    # Result with NaN values
    y_insample_nan = y_insample_orig.copy()
    y_insample_nan[-1, 0] = np.nan
    result_nan = cls_top_down(
        S=data.S, y_hat=data.S @ data.y_hat_bottom, y_insample=y_insample_nan, tags=data.tags
    )["mean"]

    np.testing.assert_allclose(result_orig, result_nan)






@pytest.mark.parametrize("method", ["forecast_proportions", "average_proportions", "proportion_averages"])
def test_middle_out_forecast_recovery(hierarchical_data, method):
    """Test MiddleOut methods can recover forecasts when time series share same proportions."""
    data = hierarchical_data

    cls_middle_out = MiddleOut(middle_level="level2", top_down_method=method)
    if cls_middle_out.insample:
        assert method in ["average_proportions", "proportion_averages"]
        result = cls_middle_out(
            S=data.S, y_hat=data.S @ data.y_hat_bottom, y_insample=data.S @ data.y_bottom, tags=data.tags
        )["mean"]
    else:
        result = cls_middle_out(S=data.S, y_hat=data.S @ data.y_hat_bottom, tags=data.tags)["mean"]

    np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


@pytest.mark.parametrize("method", ["forecast_proportions", "average_proportions", "proportion_averages"])
def test_middle_out_sparse_forecast_recovery(hierarchical_data, method):
    """Test MiddleOutSparse methods can recover forecasts."""
    data = hierarchical_data

    cls_middle_out = MiddleOutSparse(middle_level="level2", top_down_method=method)
    if cls_middle_out.insample:
        assert method in ["average_proportions", "proportion_averages"]
        result = cls_middle_out(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
    else:
        result = cls_middle_out(S=data.S, y_hat=data.S @ data.y_hat_bottom, tags=data.tags)["mean"]

    np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


@pytest.mark.parametrize("method", ["forecast_proportions", "average_proportions", "proportion_averages"])
def test_middle_out_vs_sparse_equivalence(hierarchical_data, method):
    """Test MiddleOut and MiddleOutSparse produce equivalent results."""
    data = hierarchical_data

    cls_middle_out = MiddleOut(middle_level="level2", top_down_method=method)
    cls_middle_out_sparse = MiddleOutSparse(middle_level="level2", top_down_method=method)

    if cls_middle_out.insample:
        assert method in ["average_proportions", "proportion_averages"]
        result_dense = cls_middle_out(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
        result_sparse = cls_middle_out_sparse(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
        tolerance = np.finfo(np.float64).eps
    else:
        result_dense = cls_middle_out(S=data.S, y_hat=data.S @ data.y_hat_bottom, tags=data.tags)["mean"]
        result_sparse = cls_middle_out_sparse(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            tags=data.tags,
        )["mean"]
        tolerance = np.finfo(np.float64).eps

    np.testing.assert_allclose(result_dense, result_sparse, tolerance)


@pytest.mark.parametrize("method", ["ols", "wls_struct", "wls_var", "mint_shrink", "emint"])
@pytest.mark.parametrize("nonnegative", [False, True])
def test_min_trace_forecast_recovery(hierarchical_data, method, nonnegative):
    """Test MinTrace methods can recover forecasts and nonnegative behavior."""
    data = hierarchical_data

    cls_min_trace = MinTrace(method=method, nonnegative=nonnegative)
    assert cls_min_trace.nonnegative is nonnegative

    if cls_min_trace.insample:
        assert method in ["wls_var", "mint_cov", "mint_shrink", "emint"]
        result = cls_min_trace(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            y_hat_insample=data.S @ data.y_hat_bottom_insample,
        )["mean"]
    else:
        result = cls_min_trace(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
        )["mean"]

    # EMinT without nonnegative only guarantees coherence, not exact recovery
    # (when nonnegative=True, it uses BottomUp which does have exact recovery)
    if method == "emint" and not nonnegative:
        # Check coherence instead of exact recovery
        bottom_forecasts = result[list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0])), :]
        aggregated = data.S @ bottom_forecasts
        np.testing.assert_allclose(result, aggregated, rtol=1e-10)
    else:
        np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


def test_min_trace_threading(hierarchical_data):
    """Test MinTrace produces same results with different thread counts."""
    data = hierarchical_data

    mintrace_1thr = MinTrace(method="ols", nonnegative=False, num_threads=1).fit(
        S=data.S, y_hat=data.S @ data.y_hat_bottom
    )
    mintrace_2thr = MinTrace(method="ols", nonnegative=False, num_threads=2).fit(
        S=data.S, y_hat=data.S @ data.y_hat_bottom
    )

    np.testing.assert_allclose(mintrace_1thr.y_hat, mintrace_2thr.y_hat)


@pytest.mark.parametrize("nonnegative", [False, True])
def test_min_trace_mint_cov_error(hierarchical_data, nonnegative):
    """Test MinTrace raises error for mint_cov method."""
    data = hierarchical_data

    cls_min_trace = MinTrace(method="mint_cov", nonnegative=nonnegative)
    with pytest.raises(Exception, match="min_trace.*mint_cov"):
        cls_min_trace(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            y_insample=data.S @ data.y_bottom,
            y_hat_insample=data.S @ data.y_hat_bottom_insample,
        )


def test_min_trace_shrink_covariance_stress(hierarchical_data):
    """Test MinTrace-shr covariance with different length data."""
    data = hierarchical_data

    diff_len_y_insample = data.S @ data.y_bottom
    diff_len_y_hat_insample = data.S @ data.y_hat_bottom_insample
    diff_len_y_insample[-1, :-1] = np.nan
    diff_len_y_hat_insample[-1, :-1] = np.nan

    cls_min_trace = MinTrace(method="mint_shrink")
    result_min_trace = cls_min_trace(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=diff_len_y_insample,
        y_hat_insample=diff_len_y_hat_insample,
    )

    # Test passes if no exception is raised
    assert result_min_trace is not None


@pytest.mark.parametrize("method", ["ols", "wls_struct", "wls_var", "mint_shrink", "emint"])
@pytest.mark.parametrize("nonnegative", [False, True])
def test_min_trace_levels(hierarchical_data, method, nonnegative):
    """Test MinTrace with all levels."""
    data = hierarchical_data

    cls_min_trace = MinTrace(method=method, nonnegative=nonnegative)
    result = cls_min_trace(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=data.S @ data.y_bottom,
        y_hat_insample=data.S @ data.y_hat_bottom_insample,
    )["mean"]

    # EMinT without nonnegative only guarantees coherence, not exact recovery
    # (when nonnegative=True, it uses BottomUp which does have exact recovery)
    if method == "emint" and not nonnegative:
        # Check coherence instead of exact recovery
        bottom_forecasts = result[list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0])), :]
        aggregated = data.S @ bottom_forecasts
        np.testing.assert_allclose(result, aggregated, rtol=1e-10)
    else:
        np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


@pytest.mark.parametrize("method", ["ols", "wls_struct", "wls_var"])
@pytest.mark.parametrize("nonnegative,qp", [(False, False), (True, False), (True, True)])
def test_min_trace_sparse_functionality(hierarchical_data, method, nonnegative, qp):
    """Test MinTraceSparse with non-negative heuristic and QP solutions."""
    data = hierarchical_data

    cls_min_trace = MinTraceSparse(method=method, nonnegative=nonnegative, qp=qp)
    result = cls_min_trace(
        S=sparse.csr_matrix(data.S),
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=data.S @ data.y_bottom,
        y_hat_insample=data.S @ data.y_hat_bottom_insample,
    )["mean"]

    np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


@pytest.mark.parametrize("method", ["ols", "wls_struct"])
@pytest.mark.parametrize("nonnegative", [False, True])
def test_optimal_combination_methods(hierarchical_data, method, nonnegative):
    # test nonnegative behavior
    # we should be able to recover the same forecasts
    # in this example
    cls_optimal_combination = OptimalCombination(
        method=method, nonnegative=nonnegative
    )
    np.testing.assert_allclose(
        cls_optimal_combination(
            S=hierarchical_data.S,
            y_hat=hierarchical_data.S @ hierarchical_data.y_hat_bottom,
        )["mean"],
        hierarchical_data.S @ hierarchical_data.y_hat_bottom,
    )



def test_erm_forecast_recovery(hierarchical_data):
    """Test ERM method can recover forecasts."""
    data = hierarchical_data

    cls_erm = ERM(method="reg_bu", lambda_reg=None)
    result = cls_erm(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=data.S @ data.y_bottom,
        y_hat_insample=data.S @ data.y_hat_bottom_insample,
    )["mean"]

    np.testing.assert_allclose(result, data.S @ data.y_hat_bottom)


@pytest.fixture
def interval_reconciler_args(hierarchical_data):
    """Fixture providing reconciler arguments for interval tests."""
    data = hierarchical_data
    y_base = data.S @ data.y_bottom
    y_hat_base = data.S @ data.y_hat_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample
    sigma = np.nansum((y_base - y_hat_base_insample) ** 2, axis=1) / (y_base.shape[1] - 1)
    sigma = np.sqrt(sigma)
    sigmah = sigma[:, None] * np.sqrt(
        np.vstack([np.arange(1, data.h + 1) for _ in range(y_base.shape[0])])
    )

    return dict(
        S=data.S,
        y_hat=y_hat_base,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
        sigmah=sigmah,
        level=[80, 90],
        intervals_method="normality",
        num_samples=200,
        seed=0,
        tags=data.tags,
    )


def test_bottom_up_normality_intervals(hierarchical_data, interval_reconciler_args):
    """Test bottom up normality prediction intervals recover original sigmah."""
    data = hierarchical_data

    cls_bottom_up = BottomUp()
    result = cls_bottom_up(**interval_reconciler_args)

    idx_bottom = list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0]))
    assert np.array_equal(result["sigmah"][idx_bottom], interval_reconciler_args["sigmah"][idx_bottom])


def test_bottom_up_interval_names(interval_reconciler_args):
    """Test bottom up normality interval names."""
    cls_bottom_up = BottomUp()
    bu_intervals = cls_bottom_up(**interval_reconciler_args)

    assert list(bu_intervals.keys()) == ["mean", "sigmah", "quantiles"]


def test_bottom_up_permbu_interval_names(interval_reconciler_args):
    """Test bottom up PERMBU interval names."""
    interval_reconciler_args["intervals_method"] = "permbu"
    cls_bottom_up = BottomUp()
    bu_permbu_intervals = cls_bottom_up(**interval_reconciler_args)

    assert list(bu_permbu_intervals.keys()) == ["mean", "quantiles"]


@pytest.mark.parametrize("method", ["average_proportions", "proportion_averages"])
@pytest.mark.parametrize("intervals_method", ["normality", "bootstrap", "permbu"])
def test_top_down_intervals(interval_reconciler_args, method, intervals_method):
    """Test TopDown with different interval methods."""
    interval_reconciler_args["intervals_method"] = intervals_method
    cls_top_down = TopDown(method=method)

    # Test passes if no exception is raised
    result = cls_top_down(**interval_reconciler_args)
    assert result is not None


def test_top_down_forecast_proportions_intervals(hierarchical_data):
    """Test TopDown forecast_proportions with prediction intervals."""
    data = hierarchical_data
    y_hat = data.S @ data.y_hat_bottom
    y_insample = data.S @ data.y_bottom
    y_hat_insample = data.S @ data.y_hat_bottom_insample

    cls_top_down = TopDown(method="forecast_proportions")
    result = cls_top_down(
        S=data.S,
        y_hat=y_hat,
        tags=data.tags,
        y_insample=y_insample,
        y_hat_insample=y_hat_insample,
        level=[80, 90],
        intervals_method="bootstrap",
        num_samples=100,
        seed=42,
    )

    # Check mean shape
    assert result["mean"].shape == y_hat.shape
    # Check quantiles shape: (n_series, horizon, 4) for levels [80, 90]
    assert result["quantiles"].shape == (y_hat.shape[0], y_hat.shape[1], 4)
    # Check quantiles are ordered (lower bounds < upper bounds)
    assert np.all(result["quantiles"][:, :, 0] <= result["quantiles"][:, :, 1])
    assert np.all(result["quantiles"][:, :, 2] <= result["quantiles"][:, :, 3])


def test_top_down_forecast_proportions_intervals_missing_insample(hierarchical_data):
    """Test TopDown forecast_proportions raises error when insample data missing."""
    data = hierarchical_data
    y_hat = data.S @ data.y_hat_bottom

    cls_top_down = TopDown(method="forecast_proportions")
    with pytest.raises(ValueError, match="require.*y_insample.*y_hat_insample"):
        cls_top_down(
            S=data.S,
            y_hat=y_hat,
            tags=data.tags,
            level=[80, 90],
        )


def test_top_down_sparse_forecast_proportions_intervals(hierarchical_data):
    """Test TopDownSparse forecast_proportions with prediction intervals."""
    data = hierarchical_data
    y_hat = data.S @ data.y_hat_bottom
    y_insample = data.S @ data.y_bottom
    y_hat_insample = data.S @ data.y_hat_bottom_insample

    cls_top_down = TopDownSparse(method="forecast_proportions")
    result = cls_top_down(
        S=sparse.csr_matrix(data.S),
        y_hat=y_hat,
        tags=data.tags,
        y_insample=y_insample,
        y_hat_insample=y_hat_insample,
        level=[80, 90],
        intervals_method="bootstrap",
        num_samples=100,
        seed=42,
    )

    # Check mean shape
    assert result["mean"].shape == y_hat.shape
    # Check quantiles shape: (n_series, horizon, 4) for levels [80, 90]
    assert result["quantiles"].shape == (y_hat.shape[0], y_hat.shape[1], 4)
    # Check quantiles are ordered
    assert np.all(result["quantiles"][:, :, 0] <= result["quantiles"][:, :, 1])
    assert np.all(result["quantiles"][:, :, 2] <= result["quantiles"][:, :, 3])


def test_top_down_sparse_forecast_proportions_intervals_missing_insample(hierarchical_data):
    """Test TopDownSparse forecast_proportions raises error when insample data missing."""
    data = hierarchical_data
    y_hat = data.S @ data.y_hat_bottom

    cls_top_down = TopDownSparse(method="forecast_proportions")
    with pytest.raises(ValueError, match="require.*y_insample.*y_hat_insample"):
        cls_top_down(
            S=sparse.csr_matrix(data.S),
            y_hat=y_hat,
            tags=data.tags,
            level=[80, 90],
        )


@pytest.mark.parametrize("method", ["ols", "wls_struct", "wls_var", "mint_shrink", "emint"])
@pytest.mark.parametrize("intervals_method", ["normality", "bootstrap", "permbu"])
def test_min_trace_intervals(interval_reconciler_args, method, intervals_method):
    """Test MinTrace with different interval methods."""
    interval_reconciler_args["intervals_method"] = intervals_method
    cls_min_trace = MinTrace(method=method, nonnegative=False)

    # Test passes if no exception is raised
    result = cls_min_trace(**interval_reconciler_args)
    assert result is not None


@pytest.mark.parametrize("method", ["ols", "wls_struct"])
@pytest.mark.parametrize("intervals_method", ["normality", "bootstrap", "permbu"])
def test_optimal_combination_intervals(interval_reconciler_args, method, intervals_method):
    """Test OptimalCombination with different interval methods."""
    interval_reconciler_args["intervals_method"] = intervals_method
    cls_optimal_combination = OptimalCombination(method=method, nonnegative=False)

    # Test passes if no exception is raised (only for non-negative=False)
    result = cls_optimal_combination(**interval_reconciler_args)
    assert result is not None


@pytest.mark.parametrize("intervals_method", ["normality", "bootstrap", "permbu"])
def test_erm_intervals(interval_reconciler_args, intervals_method):
    """Test ERM with different interval methods."""
    interval_reconciler_args["intervals_method"] = intervals_method
    cls_erm = ERM(method="reg_bu", lambda_reg=None)

    # Test passes if no exception is raised
    result = cls_erm(**interval_reconciler_args)
    assert result is not None


@pytest.mark.parametrize("intervals_method", ["normality", "bootstrap", "permbu"])
def test_coherent_sample_shapes(hierarchical_data, intervals_method):
    """Test coherent sample shapes are consistent across interval methods."""
    data = hierarchical_data
    y_base = data.S @ data.y_bottom
    y_hat_base = data.S @ data.y_hat_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample
    sigma = np.nansum((y_base - y_hat_base_insample) ** 2, axis=1) / (y_base.shape[1] - 1)
    sigma = np.sqrt(sigma)
    sigmah = sigma[:, None] * np.sqrt(
        np.vstack([np.arange(1, data.h + 1) for _ in range(y_base.shape[0])])
    )

    reconciler_args = dict(
        S=data.S,
        y_hat=y_hat_base,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
        sigmah=sigmah,
        intervals_method=intervals_method,
        tags=data.tags,
    )

    cls_bottom_up = BottomUp()
    cls_bottom_up.fit(**reconciler_args)
    coherent_samples = cls_bottom_up.sample(num_samples=100)

    # Store shape for comparison in a class variable
    if not hasattr(test_coherent_sample_shapes, 'shapes'):
        test_coherent_sample_shapes.shapes = []
    test_coherent_sample_shapes.shapes.append(coherent_samples.shape)

    # If we have multiple shapes, compare them
    if len(test_coherent_sample_shapes.shapes) > 1:
        assert test_coherent_sample_shapes.shapes[0] == test_coherent_sample_shapes.shapes[-1]


def test_mintrace_nonnegative_samples_use_constrained_forecasts(hierarchical_data):
    """Test that MinTrace nonnegative samples are based on constrained forecasts.

    This is a regression test for the bug where _get_sampler() was called with
    the original y_hat instead of the nonnegative-constrained self.y_hat,
    causing samples to potentially contain negative values even when
    nonnegative=True.
    """
    data = hierarchical_data

    # Create base forecasts with NEGATIVE values to trigger nonnegative constraint
    y_hat_base = data.S @ data.y_hat_bottom
    # Make some forecasts negative - use large negative values to make the bug obvious
    y_hat_with_negatives = y_hat_base.copy()
    y_hat_with_negatives[0, :] = -10.0  # Make total negative
    y_hat_with_negatives[1, :] = -5.0   # Make group 1 negative

    # Create in-sample data
    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample

    # Compute sigmah - use small values to reduce sampling variance
    sigmah = np.ones((data.S.shape[0], data.h)) * 0.5

    # Use MinTrace with nonnegative=True and normality intervals
    # (bootstrap/permbu are not compatible with nonnegative)
    cls_min_trace = MinTrace(method="ols", nonnegative=True)

    result = cls_min_trace.fit_predict(
        S=data.S,
        y_hat=y_hat_with_negatives,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
        sigmah=sigmah,
        level=[80, 90],
        intervals_method="normality",
        num_samples=200,
        seed=42,
        tags=data.tags,
    )

    # Verify the mean reconciled forecasts are non-negative
    assert np.all(result["mean"] >= 0), "Mean forecasts should be non-negative"

    # KEY ASSERTION: Verify the sampler was initialized with the nonnegative-constrained y_hat
    # This is the core fix - previously the sampler was initialized with original y_hat
    assert cls_min_trace.sampler is not None, "Sampler should be initialized"
    np.testing.assert_array_equal(
        cls_min_trace.sampler.y_hat,
        cls_min_trace.y_hat,
        err_msg="Sampler y_hat should match the nonnegative-constrained y_hat"
    )

    # Verify sampler y_hat is non-negative (since we used nonnegative=True)
    assert np.all(cls_min_trace.sampler.y_hat >= -1e-6), (
        "Sampler y_hat should be non-negative when nonnegative=True"
    )

    # Generate samples - if the bug existed, samples would be centered around
    # the original negative values (-10, -5) instead of the constrained values
    samples = cls_min_trace.sample(num_samples=500)
    sample_means = np.mean(samples, axis=2)

    # With the bug fixed, sample means should be close to the non-negative reconciled mean.
    # Tolerance explanation: With sigmah=0.5 and 500 samples, the standard error of the mean
    # is approximately sigmah/sqrt(500) ≈ 0.022. We use atol=0.3 to account for the
    # coherent covariance structure which can increase variance.
    np.testing.assert_allclose(
        sample_means,
        result["mean"],
        atol=0.3,
        err_msg="Sample means should be close to non-negative reconciled forecasts"
    )


def test_mintrace_nonnegative_without_intervals(hierarchical_data):
    """Test MinTrace with nonnegative=True but without probabilistic intervals."""
    data = hierarchical_data

    y_hat_base = data.S @ data.y_hat_bottom
    y_hat_with_negatives = y_hat_base.copy()
    y_hat_with_negatives[0, :] = -10.0

    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample

    cls_min_trace = MinTrace(method="ols", nonnegative=True)

    # Should work without intervals (level=None, intervals_method=None)
    result = cls_min_trace.fit_predict(
        S=data.S,
        y_hat=y_hat_with_negatives,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
        sigmah=None,
        level=None,
        intervals_method=None,
        num_samples=None,
        seed=None,
        tags=data.tags,
    )

    # Verify the mean reconciled forecasts are non-negative
    assert np.all(result["mean"] >= 0), "Mean forecasts should be non-negative"
    # Should not have quantiles since level=None
    assert "quantiles" not in result or result["quantiles"] is None


@pytest.mark.parametrize("method", ["ols", "wls_struct"])
def test_mintrace_sparse_nonnegative_sampler_initialization(hierarchical_data, method):
    """Test that MinTraceSparse with nonnegative=True correctly initializes sampler."""
    data = hierarchical_data

    # Convert S to sparse matrix as required by MinTraceSparse
    S_sparse = sparse.csr_matrix(data.S)

    y_hat_base = data.S @ data.y_hat_bottom
    y_hat_with_negatives = y_hat_base.copy()
    y_hat_with_negatives[0, :] = -10.0

    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample
    sigmah = np.ones((data.S.shape[0], data.h)) * 0.5

    cls_min_trace = MinTraceSparse(method=method, nonnegative=True)

    result = cls_min_trace.fit_predict(
        S=S_sparse,
        y_hat=y_hat_with_negatives,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
        sigmah=sigmah,
        level=[80, 90],
        intervals_method="normality",
        num_samples=200,
        seed=42,
        tags=data.tags,
    )

    # Verify the mean reconciled forecasts are non-negative
    assert np.all(result["mean"] >= 0), "Mean forecasts should be non-negative"

    # Verify sampler uses the constrained y_hat
    assert cls_min_trace.sampler is not None
    np.testing.assert_array_equal(
        cls_min_trace.sampler.y_hat,
        cls_min_trace.y_hat,
        err_msg="MinTraceSparse sampler y_hat should match nonnegative-constrained y_hat"
    )


@pytest.mark.parametrize("intervals_method", ["bootstrap", "permbu"])
def test_mintrace_nonnegative_raises_on_bootstrap_permbu(hierarchical_data, intervals_method):
    """Test that MinTrace with nonnegative=True raises error for bootstrap/permbu.

    Nonnegative reconciliation is not compatible with bootstrap or permbu probabilistic
    forecasts because these methods generate samples from historical residuals which
    do not respect the nonnegative constraint applied during reconciliation.
    """
    data = hierarchical_data

    y_hat_base = data.S @ data.y_hat_bottom
    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample
    sigmah = np.ones((data.S.shape[0], data.h)) * 0.5

    cls_min_trace = MinTrace(method="ols", nonnegative=True)

    with pytest.raises(ValueError, match="nonnegative reconciliation is not compatible"):
        cls_min_trace.fit_predict(
            S=data.S,
            y_hat=y_hat_base,
            y_insample=y_base,
            y_hat_insample=y_hat_base_insample,
            sigmah=sigmah,
            level=[90],
            intervals_method=intervals_method,
            num_samples=200,
            seed=42,
            tags=data.tags,
            )


@pytest.mark.parametrize("nonnegative", [False, True])
def test_emint_forecast_reconciliation(hierarchical_data, nonnegative):
    """Test EMinT method produces coherent forecasts."""
    data = hierarchical_data

    cls_emint = MinTrace(method='emint', nonnegative=nonnegative)
    result = cls_emint(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=data.S @ data.y_bottom,
        y_hat_insample=data.S @ data.y_hat_bottom_insample,
    )["mean"]

    # Check that result is coherent (satisfies aggregation constraints)
    # Bottom level forecasts should aggregate to upper levels
    bottom_forecasts = result[list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0])), :]
    aggregated = data.S @ bottom_forecasts

    np.testing.assert_allclose(result, aggregated, rtol=1e-10)

    # If nonnegative=True, verify all forecasts are non-negative
    if nonnegative:
        assert np.all(result >= -1e-6), "Forecasts should be non-negative when nonnegative=True"


def test_emint_fit_predict_pattern(hierarchical_data):
    """Test EMinT with fit -> predict pattern."""
    data = hierarchical_data

    cls_emint = MinTrace(method='emint')
    cls_emint.fit(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=data.S @ data.y_bottom,
        y_hat_insample=data.S @ data.y_hat_bottom_insample,
    )
    result = cls_emint.predict(S=data.S, y_hat=data.S @ data.y_hat_bottom)["mean"]

    # Check coherence
    bottom_forecasts = result[list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0])), :]
    aggregated = data.S @ bottom_forecasts

    np.testing.assert_allclose(result, aggregated, rtol=1e-10)


def test_emint_requires_insample_data(hierarchical_data):
    """Test EMinT raises error when insample data is missing."""
    data = hierarchical_data

    cls_emint = MinTrace(method='emint')

    with pytest.raises(ValueError, match="insample predictions and insample values"):
        cls_emint(
            S=data.S,
            y_hat=data.S @ data.y_hat_bottom,
            )


@pytest.mark.parametrize("method_class,kwargs", [
    (MinTrace, {"method": "emint"}),
])
def test_insample_methods_coherence(hierarchical_data, method_class, kwargs):
    """Test that EMinT produces coherent forecasts."""
    data = hierarchical_data

    reconciler = method_class(**kwargs)
    result = reconciler(
        S=data.S,
        y_hat=data.S @ data.y_hat_bottom,
        y_insample=data.S @ data.y_bottom,
        y_hat_insample=data.S @ data.y_hat_bottom_insample,
    )["mean"]

    # Verify coherence: S @ bottom = all
    bottom_forecasts = result[list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0])), :]
    aggregated = data.S @ bottom_forecasts

    np.testing.assert_allclose(result, aggregated, rtol=1e-10)

    # Verify shape
    assert result.shape == (data.S.shape[0], data.h)


def test_emint_nonnegative_with_negative_forecasts(hierarchical_data):
    """Test EMinT with nonnegative=True handles negative base forecasts."""
    data = hierarchical_data

    # Create base forecasts with NEGATIVE values
    y_hat_base = data.S @ data.y_hat_bottom
    y_hat_with_negatives = y_hat_base.copy()
    y_hat_with_negatives[0, :] = -10.0  # Make total negative
    y_hat_with_negatives[1, :] = -5.0   # Make group 1 negative

    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample

    cls_emint = MinTrace(method='emint', nonnegative=True)

    result = cls_emint.fit_predict(
        S=data.S,
        y_hat=y_hat_with_negatives,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
    )

    # Verify the mean reconciled forecasts are non-negative
    assert np.all(result["mean"] >= -1e-6), "Mean forecasts should be non-negative"

    # Verify coherence
    bottom_forecasts = result["mean"][list(range(data.S.shape[0] - data.S.shape[1], data.S.shape[0])), :]
    aggregated = data.S @ bottom_forecasts
    np.testing.assert_allclose(result["mean"], aggregated, rtol=1e-10)


def test_emint_nonnegative_sampler_initialization(hierarchical_data):
    """Test that EMinT with nonnegative=True correctly initializes sampler with constrained forecasts."""
    data = hierarchical_data

    y_hat_base = data.S @ data.y_hat_bottom
    y_hat_with_negatives = y_hat_base.copy()
    y_hat_with_negatives[0, :] = -10.0

    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample
    sigmah = np.ones((data.S.shape[0], data.h)) * 0.5

    cls_emint = MinTrace(method='emint', nonnegative=True)

    result = cls_emint.fit_predict(
        S=data.S,
        y_hat=y_hat_with_negatives,
        y_insample=y_base,
        y_hat_insample=y_hat_base_insample,
        sigmah=sigmah,
        level=[80, 90],
        intervals_method="normality",
        num_samples=200,
        seed=42,
        tags=data.tags,
    )

    # Verify the mean reconciled forecasts are non-negative
    assert np.all(result["mean"] >= -1e-6), "Mean forecasts should be non-negative"

    # Verify the sampler was initialized with the nonnegative-constrained y_hat
    np.testing.assert_array_equal(
        cls_emint.sampler.y_hat,
        cls_emint.y_hat,
        err_msg="Sampler y_hat should match the nonnegative-constrained y_hat"
    )

    # Verify sampler y_hat is non-negative
    assert np.all(cls_emint.sampler.y_hat >= -1e-6), (
        "Sampler y_hat should be non-negative when nonnegative=True"
    )


@pytest.mark.parametrize("intervals_method", ["bootstrap", "permbu"])
def test_emint_nonnegative_raises_on_bootstrap_permbu(hierarchical_data, intervals_method):
    """Test that EMinT with nonnegative=True raises error for bootstrap/permbu."""
    data = hierarchical_data

    y_hat_base = data.S @ data.y_hat_bottom
    y_base = data.S @ data.y_bottom
    y_hat_base_insample = data.S @ data.y_hat_bottom_insample
    sigmah = np.ones((data.S.shape[0], data.h)) * 0.5

    cls_emint = MinTrace(method='emint', nonnegative=True)

    with pytest.raises(ValueError, match="nonnegative reconciliation is not compatible"):
        cls_emint.fit_predict(
            S=data.S,
            y_hat=y_hat_base,
            y_insample=y_base,
            y_hat_insample=y_hat_base_insample,
            sigmah=sigmah,
            level=[90],
            intervals_method=intervals_method,
            num_samples=200,
            seed=42,
            tags=data.tags,
            )
